Attribute VB_Name = "mdlQFunction"
'state is encoded as an integer
' <state> = <location> + 25 * < buttons>
' <buttons> = 2^[0 7] for each button in each floor
' <location> = elevator1 + 5 * elevator2
Public state As Integer
'actions are coded as:
' 0: NOP el1    0:NOP el2
' 1: up el1     4:up el2
' 2: down el1   8:down el2
' 3: stop el1   12:stop el2
Public QFunction(6424, 15) As Single
'Q Learning constants
Const alpha = 0.5
Const lambda = 0.9
Public epsilon As Single


Public Sub initQFunction()
For i = 0 To 6424
    For j = 0 To 2
        QFunction(i, j) = 0
    Next j
Next i
epsilon = 1

End Sub

Public Function CalculateQAction(state As Integer, freeElevator As Integer) As Integer
'free elevator is coded
'1: el1 2: el2 3: el1 and el2

Dim maxValue As Single

'randomize to see if greedy on Q Function or random
Dim greedy As Boolean
greedy = Rnd() > epsilon
Dim secondElevatorAction As Integer

'if only elevator 1
If freeElevator = 1 Then
    'get the other elevator action
    secondElevatorAction = decodeAction(2)
    maxValue = QFunction(state, 1)
    CalculateQAction = 3 + 4 * secondElevatorAction
    If greedy = True Then
        For i = 1 To 3
            If maxValue < QFunction(state, i) Then
                maxValue = QFunction(state, i)
                CalculateQAction = i + 4 * secondElevatorAction
            End If
        Next i
    Else
        CalculateQAction = Int(Rnd() * 3 + 1) + 4 * secondElevatorAction
    End If
'if only elevator 2
ElseIf freeElevator = 2 Then
    'get the other elevator action
    secondElevatorAction = decodeAction(1)
    maxValue = QFunction(state, 4)
    CalculateQAction = 3 + secondElevatorAction
    If greedy = True Then
        For i = 4 To 12 Step 4
            If maxValue < QFunction(state, i) Then
                maxValue = QFunction(state, i)
                CalculateQAction = i + secondElevatorAction
            End If
        Next i
    Else
        CalculateQAction = Int(Rnd() * 3 + 1) * 4 + secondElevatorAction
    End If
'if both elevators
ElseIf freeElevator = 3 Then
    maxValue = QFunction(state, 5)
    CalculateQAction = 15
    If greedy = True Then
        For i = 5 To 15
            If maxValue < QFunction(state, i) Then
                maxValue = QFunction(state, i)
                CalculateQAction = i
            End If
        Next i
    Else
        CalculateQAction = Int(Rnd() * 11 + 5)
    End If
End If

End Function

Public Sub updateQFunction(curState As Integer, newState As Integer, curAction As Integer, curReward As Double)

'find max Q
Dim maxQ As Single
maxQ = QFunction(newState, 0)
For i = 0 To 15
    If maxQ < QFunction(newState, i) Then maxQ = QFunction(newState, i)
Next i

QFunction(curState, curAction) = (1 - alpha) * QFunction(curState, curAction) + alpha * (curReward + lambda * maxQ)

End Sub
Private Function selectStochasticValue(valueArray() As Single, arraySize As Integer) As Integer

Dim sum, aggregator, randomNumber As Single
Dim index As Integer
Dim probabilityArray(15) As Single
sum = 0
aggregator = 0

'generate a normalized array of probabilities
For index = 0 To arraySize - 1
    probabilityArray(index) = Exp(valueArray(index))
    sum = sum + probabilityArray(index)
Next index
'normalize to a [0 1] range
If sum <> 0 Then
    For index = 0 To arraySize - 1
        probabilityArray(index) = probabilityArray(index) / sum
    Next index
    'select action randomly
    randomNumber = Rnd()
    For index = 0 To arraySize - 1
        If randomNumber < (probabilityArray(index) + aggregator) Then
            selectStochasticValue = index
            index = arraySize
        Else
            aggregator = aggregator + probabilityArray(index)
        End If
    Next index
Else 'Q values are too big to give non zero exponents
    'decrease them by a factor of square root
    For index = 0 To arraySize - 1
        valueArray(index) = -1 * Sqr(-1 * valueArray(index))
    Next index
    'call recursively
    selectStochasticValue = selectStochasticValue(valueArray, arraySize)
End If




End Function

Public Function newCalculateQAction(state As Integer, freeElevator As Integer) As Integer
'free elevator is coded
'1: el1 2: el2 3: el1 and el2

Dim secondElevatorAction As Integer
Dim valueArray(15) As Single
Dim myAction, index As Integer

'if only elevator 1
If freeElevator = 1 Then
    'get the other elevator action
    secondElevatorAction = decodeAction(2)
    'create array with Q values of possible actions
    For index = 0 To 3
        valueArray(index) = QFunction(state, index)
    Next index
    myAction = selectStochasticValue(valueArray, 4)
    newCalculateQAction = myAction + 4 * secondElevatorAction
'if only elevator 2
ElseIf freeElevator = 2 Then
    'get the other elevator action
    secondElevatorAction = decodeAction(1)
    For index = 0 To 12 Step 4
        valueArray(index / 4) = QFunction(state, index)
    Next index
    myAction = selectStochasticValue(valueArray, 4)
    newCalculateQAction = myAction * 4 + secondElevatorAction
'if both elevators
ElseIf freeElevator = 3 Then
    For index = 0 To 15
        valueArray(index) = QFunction(state, index)
    Next index
    myAction = selectStochasticValue(valueArray, 16)
    newCalculateQAction = myAction
End If

End Function
