Attribute VB_Name = "mdlSimulationControl"
Public elevatorButtons(4, 2) As Boolean
'location of elevator (ground = floor 0)
Public floor(2) As Integer
'direction is coded
'1: up 0: down
Public direction(2) As Integer
'people queue is coded in four fields
'1st field: floor the person is waiting on
'2nd field: arrival time of person
'3rd field: destination floor
'4th field: direction to destination
Public peopleQueue(1000, 4) As Long
'action is coded
'1: up 2: down 3: stop
Public action As Integer
'is each elevator in a decision point
Public decisionPoint(2) As Boolean
'couting tics from 0
Public currentTime As Long
'immediate reward at each time step
Public reward As Double
'average waiting time
Public avgWait As Single
Public numberOFPeople As Long
'the mu factor of the poisson appearence if people
Const mu = 0.5
Option Explicit

Public Sub initSimulation()
Dim index As Integer
Randomize Timer
currentTime = 0

For index = 1 To 1000
    peopleQueue(index, 1) = -1
Next index

For index = 0 To 4
    elevatorButtons(index, 1) = False
    elevatorButtons(index, 2) = False
Next index

decisionPoint(1) = False
decisionPoint(2) = False
floor(1) = 0
floor(2) = 3
direction(1) = 1
direction(2) = 0
reward = 0
avgWait = 0
numberOFPeople = 0
End Sub

Public Function calculateHeuristicAction() As Integer

calculateHeuristicAction = 0
'ret is coded
'0:NOP 1:up 2:down 3:stop
Dim ret(2) As Integer
Dim elevator, peopleIndex, buttonIndex As Integer
Dim buttonsPressed As Boolean

'for each elevator
For elevator = 1 To 2
    ret(elevator) = 0
    For peopleIndex = 1 To 1000
        'if someone is in the same floor and same direction
        If peopleQueue(peopleIndex, 1) = floor(elevator) _
        And peopleQueue(peopleIndex, 4) = direction(elevator) Then
           If direction(elevator) = 1 Then
            ret(elevator) = 1
            ElseIf direction(elevator) = 0 Then
             ret(elevator) = 2
            End If
            
        End If
    Next peopleIndex
        
        'if someone wants to get off the elevator at this floor
        
        If elevatorButtons(floor(elevator), elevator) = True Then
          If direction(elevator) = 1 Then
            ret(elevator) = 1
            ElseIf direction(elevator) = 0 Then
             ret(elevator) = 2
            End If
          End If
    
        'if didn't stop
        If ret(elevator) = 0 Then
            'if no more buttons pressed
            buttonsPressed = False
            For buttonIndex = 0 To 4
                If elevatorButtons(buttonIndex, elevator) = True Then buttonsPressed = True
            Next buttonIndex
            If buttonsPressed = True Then
                If direction(elevator) = 1 Then
                   'go up
                    ret(elevator) = 1
                Else
                    'go down
                    ret(elevator) = 2
                End If
            End If
        End If
        
        'if no orders were issued
        If ret(elevator) = 0 Then
            decisionPoint(elevator) = True
        End If
        
Next elevator

'code the action of both elevators
calculateHeuristicAction = ret(1) + 4 * ret(2)

End Function

Public Sub act()
 Dim elevator As Integer
 Dim peopleIndex, newAction As Integer
 
 For elevator = 1 To 2
    newAction = decodeAction(elevator)
    
    If newAction = 1 Then
        direction(elevator) = 1
    ElseIf newAction = 2 Then
        direction(elevator) = 0
    End If
    
     
    For peopleIndex = 1 To 1000
        'if someone is in the same floor and same direction
         If peopleQueue(peopleIndex, 1) = floor(elevator) _
         And peopleQueue(peopleIndex, 4) = direction(elevator) Then
            'person gets into elevator
            peopleQueue(peopleIndex, 1) = -1
            'press buttons in elevator for destination
            elevatorButtons(peopleQueue(peopleIndex, 3), elevator) = True
            'add the waiting time to the average waiting time
            avgWait = (avgWait * numberOFPeople + (currentTime - peopleQueue(peopleIndex, 2))) / (numberOFPeople + 1)
            numberOFPeople = numberOFPeople + 1
          End If
    Next peopleIndex
         
        'if someone wants to get off the elevator at this floor
        If elevatorButtons(floor(elevator), elevator) = True Then
            'unset the button in the elevator
            elevatorButtons(floor(elevator), elevator) = False
        End If

    'if moving up
    If newAction = 1 And floor(elevator) < 4 Then
        floor(elevator) = floor(elevator) + 1
    'if moving down
    ElseIf newAction = 2 And floor(elevator) > 0 Then
        floor(elevator) = floor(elevator) - 1
    End If
    If floor(elevator) = 4 Then
        direction(elevator) = 0
    'if moving down
    ElseIf floor(elevator) = 0 Then
        direction(elevator) = 1
    End If
    
 Next elevator

'status report
frmSimulation.lblReport.Caption = frmSimulation.lblReport.Caption & "average wait: " & CStr(avgWait) & " seconds" & vbNewLine

End Sub
Public Function calculateState() As Integer

Dim buttons(7) As Boolean
Dim i As Integer

For i = 0 To 7
    buttons(i) = False
Next i
'check for each person which button they pressed
For i = 1 To 1000
    If peopleQueue(i, 1) <> -1 Then
        Dim buttonIndex As Integer
        buttonIndex = 2 * peopleQueue(i, 1) - 1 + peopleQueue(i, 4)
        buttons(buttonIndex) = True
    End If
Next i
'code buttons
Dim buttonCode As Integer
For i = 0 To 7
    If buttons(i) = True Then buttonCode = buttonCode + 2 ^ i
Next i
'code elevator location
Dim locationCode
locationCode = floor(1) + 5 * floor(2)

'calculate state
calculateState = locationCode + 25 * buttonCode

End Function

Public Sub addPeople()

Dim number, curfloor, lobbyIndex, index, destination, dire As Integer
Dim prob As Single

'probability that a person will appear this tick
prob = Exp(-1 * mu) * mu
number = 0

'add people on floors, going down
For curfloor = 1 To 4
    If Rnd() < prob Then
        destination = Int(Rnd() * 4 + 1)
        If destination = curfloor Then
            destination = 0
        End If
        If destination < curfloor Then
            dire = 0
        Else
            dire = 1
        End If
        For index = 1 To 1000
            If peopleQueue(index, 1) = -1 Then
                peopleQueue(index, 1) = curfloor
                peopleQueue(index, 2) = currentTime
                peopleQueue(index, 3) = destination
                peopleQueue(index, 4) = dire
                index = 1000
            End If
        Next index
        number = number + 1
    End If
Next curfloor

'add the same number of people in the lobby
For lobbyIndex = 1 To Int(number / 2)
    destination = Int(Rnd() * 4 + 1)
    For index = 1 To 1000
        If peopleQueue(index, 1) = -1 Then
            peopleQueue(index, 1) = 0
            peopleQueue(index, 2) = currentTime
            peopleQueue(index, 3) = destination
            peopleQueue(index, 4) = 1
            index = 1000
        End If
    Next index
Next lobbyIndex

End Sub

Public Function calculateReward() As Double
calculateReward = 0
Dim index As Integer

'loop over all people in queue
For index = 1 To 1000
    'if this is a waiting person
    If peopleQueue(index, 1) <> -1 Then
        'add the squared time to the reward
        calculateReward = calculateReward - _
        (peopleQueue(index, 2) - currentTime) ^ 2
    End If
Next index

'status report
frmSimulation.lblReport.Caption = frmSimulation.lblReport.Caption & "Reward: " & calculateReward & vbNewLine

End Function


Public Sub updateGUI()

Dim index As Integer, elevatorAction As Integer

'set elevator location
For index = 1 To 2
    elevatorAction = decodeAction(index)
    'if moving up
    If elevatorAction = 1 Then
        Call moveElevator(index, 1)
    'if moving down
    ElseIf elevatorAction = 2 Then
        Call moveElevator(index, 0)
    End If
Next index
    
'set number of people in each floor and direction
Dim totalUpPeople(4) As Integer, totalDownPeople(4) As Integer
For index = 1 To 1000
    If peopleQueue(index, 1) <> -1 Then
        If peopleQueue(index, 4) = 1 Then
            totalUpPeople(peopleQueue(index, 1)) = totalUpPeople(peopleQueue(index, 1)) + 1
        ElseIf peopleQueue(index, 4) = 0 Then
            totalDownPeople(peopleQueue(index, 1)) = totalDownPeople(peopleQueue(index, 1)) + 1
        End If
    End If
Next index
 
'display people in GUI
For index = 0 To 4
    Call setNumberOfDowngoingPeople(index, totalDownPeople(index))
    Call setNumberOfUpgoingPeople(index, totalUpPeople(index))
Next index

End Sub

Public Function decodeAction(elevator As Integer) As Integer
'elevator is the elevator number (not both !)
'if elevator is invalid, return -1

If elevator = 1 Then
    decodeAction = action Mod 4
ElseIf elevator = 2 Then
    decodeAction = Int(action / 4)
Else
    decodeAction = -1
End If

End Function
